package ru.landgrafhomyak.serdha.api.v0.ddl.types

import kotlin.jvm.JvmField
import kotlin.jvm.JvmName
import kotlinx.datetime.Instant
import kotlinx.datetime.LocalDate
import kotlinx.datetime.LocalDateTime
import kotlinx.datetime.LocalTime
import ru.landgrafhomyak.serdha.api.v0.ddl.table.Table

/**
 * Descriptor of a database type.
 *
 * @param RuntimeType Type in programming language to which database type is converted.
 */
@Suppress("ClassName", "RemoveRedundantQualifierName")
public interface ColumnType<@Suppress("unused") RuntimeType> {
	/**
	 * Name of database type for debugging and errors.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("name")
	public val name: String

	/**
	 * Provider of database types descriptors.
	 */
	@Suppress("FunctionName", "PropertyName")
	public interface Builder {
		/**
		 * Type of internal row id. Can't be casted to types like integers or pointers.
		 * @param TableUserExtension User's extension type for target table. For static type checking.
		 * @param table Descriptor of table whose row_id type will be obtained.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@JvmName("ROW_ID")
		public fun <TableUserExtension : Any> ROW_ID(table: Table<TableUserExtension, *>): ColumnType<RowId<TableUserExtension>>

		/**
		 * Descriptor of a boolean type. Fields of this type can contain only 2 values: `true` and `false`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("BOOLEAN")
		public val BOOLEAN: BOOLEAN

		/**
		 * Descriptor of a signed integer type with size at least 8 bits. Can contain values in range `-128..127`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_S8")
		public val INT_S8: INT_S8

		/**
		 * Descriptor of an unsigned integer type with size at least 8 bits. Can contain values in range `0..255`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_U8")
		public val INT_U8: INT_U8

		/**
		 * Descriptor of a signed integer type with size at least 16 bits. Can contain values in range `-32768..32767`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_S16")
		public val INT_S16: INT_S16

		/**
		 * Descriptor of an unsigned integer type with size at least 16 bits. Can contain values in range `0..65535`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_U16")
		public val INT_U16: INT_U16

		/**
		 * Descriptor of a signed integer type with size at least 32 bits. Can contain values in range `-2147483648..2147483647`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_S32")
		public val INT_S32: INT_S32

		/**
		 * Descriptor of an unsigned integer type with size at least 32 bits. Can contain values in range `0..4294967295`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_U32")
		public val INT_U32: INT_U32

		/**
		 * Descriptor of a signed integer type with size at least 64 bits. Can contain values in range `-9223372036854775808..9223372036854775807`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_S64")
		public val INT_S64: INT_S64

		/**
		 * Descriptor of an unsigned integer type with size at least 64 bits. Can contain values in range `0..18446744073709551615`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INT_U64")
		public val INT_U64: INT_U64

		/**
		 * Descriptor of a local date type in some format like `(year, month, day)`. Day component can contain values in `1..31`, month in `1..12` and year in `-32768..32767`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("LOCAL_DATE")
		public val LOCAL_DATE: LOCAL_DATE

		/**
		 * Descriptor of a local time type in some format like `(hour, minute, second)`. Hour component can contain values in `0..23`, minute in `0..59` and second in `0..59`.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("LOCAL_TIME")
		public val LOCAL_TIME: LOCAL_TIME

		/**
		 * Descriptor of union of types [LOCAL_DATE][ColumnType.Builder.LOCAL_DATE] and [LOCAL_TIME][ColumnType.Builder.LOCAL_TIME].
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("LOCAL_DATETIME")
		public val LOCAL_DATETIME: LOCAL_DATETIME

		/**
		 * Descriptor of a timestamp type. Can contain values in `-9223372036854775808..9223372036854775807`
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("INSTANT")
		public val INSTANT: INSTANT

		/**
		 * Descriptor of a single-precision floating-point number with size at least 32 bits. See [IEEE 754](https://en.wikipedia.org/wiki/IEEE_754).
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("FLOAT_32")
		public val FLOAT_32: FLOAT_32

		/**
		 * Descriptor of a double-precision floating-point number with size at least 64 bits. See [IEEE 754](https://en.wikipedia.org/wiki/IEEE_754).
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("FLOAT_64")
		public val FLOAT_64: FLOAT_64

		//todo
		public val STRING: STRING

		//todo
		public fun STRING(size: UInt): STRING

		//todo
		public val BYTE_ARRAY: BYTE_ARRAY

		//todo
		public fun BYTE_ARRAY(size: UInt): BYTE_ARRAY


		/**
		 * Returns descriptor of a new type, which can contain all values of [original type][notNullType] or `null` value.
		 *
		 * Descriptors of nullable types are returned without any modifications.
		 */
		@Suppress("INAPPLICABLE_JVM_NAME")
		@JvmName("nullableOf")
		public fun <R : Any, D : ColumnType<R>> nullableOf(notNullType: D): Nullable<R, D>
	}

	// todo
	public abstract class _VirtualType<R : Any>(@Suppress("MemberVisibilityCanBePrivate") @JvmField public val wraps: ColumnType<R>)

	/**
	 * Descriptor of a nullable database type for static type-checking.
	 * @see ColumnType.Builder.nullableOf
	 */
	public interface Nullable<RuntimeType : Any, @Suppress("unused") DatabaseType : ColumnType<RuntimeType>> : ColumnType<RuntimeType?>

	/**
	 * Descriptor of an internal row id type for static type-checking.
	 * @param TargetTableUserWrapper  User's extension type for target table. For static type checking.
	 * @see ColumnType.Builder.ROW_ID
	 */
	public interface ROW_ID<TargetTableUserWrapper: Any> : ColumnType<RowId<TargetTableUserWrapper>>

	/**
	 * Descriptor of a boolean type for static type-checking.
	 * @see ColumnType.Builder.BOOLEAN
	 */
	public interface BOOLEAN : ColumnType<Boolean>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_S8
	 */
	public interface INT_S8 : ColumnType<Byte>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_U8
	 */
	public interface INT_U8 : ColumnType<UByte>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_S16
	 */
	public interface INT_S16 : ColumnType<Short>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_U16
	 */
	public interface INT_U16 : ColumnType<UShort>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_S32
	 */
	public interface INT_S32 : ColumnType<Int>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_U32
	 */
	public interface INT_U32 : ColumnType<UInt>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_S64
	 */
	public interface INT_S64 : ColumnType<Long>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see ColumnType.Builder.INT_U64
	 */
	public interface INT_U64 : ColumnType<ULong>

	/**
	 * Descriptor of a date type for static type-checking.
	 * @see ColumnType.Builder.LOCAL_DATE
	 */
	public interface LOCAL_DATE : ColumnType<LocalDate>

	/**
	 * Descriptor of a time type for static type-checking.
	 * @see ColumnType.Builder.LOCAL_TIME
	 */
	public interface LOCAL_TIME : ColumnType<LocalTime>

	/**
	 * Descriptor of a date-time type for static type-checking.
	 * @see ColumnType.Builder.LOCAL_DATETIME
	 */
	public interface LOCAL_DATETIME : ColumnType<LocalDateTime>

	/**
	 * Descriptor of a date-time type for static type-checking.
	 * @see ColumnType.Builder.INSTANT
	 */
	public interface INSTANT : ColumnType<Instant>

	/**
	 * Descriptor of a floating-point number type for static type-checking.
	 * @see ColumnType.Builder.FLOAT_32
	 */
	public interface FLOAT_32 : ColumnType<Float>

	/**
	 * Descriptor of a floating-point number type for static type-checking.
	 * @see ColumnType.Builder.FLOAT_64
	 */
	public interface FLOAT_64 : ColumnType<Double>

	/**
	 * Descriptor of a string type for static type-checking.
	 * @see ColumnType.Builder.STRING
	 */
	public interface STRING : ColumnType<String>

	/**
	 * Descriptor of a raw binary data type for static type-checking.
	 * @see ColumnType.Builder.BYTE_ARRAY
	 */
	@OptIn(ExperimentalUnsignedTypes::class)
	public interface BYTE_ARRAY : ColumnType<UByteArray>
}
